import os.path
from typing import Union

import torch
from torch import nn
from torch.nn import DataParallel
from torch.nn.parallel import DistributedDataParallel
from torch.optim import Optimizer

from jbag.log import logger

MODEL = 'model'
OPTIMIZER = 'optimizer'


def get_unwrapped_model(model: nn.Module):
    if isinstance(model, (DataParallel, DistributedDataParallel)):
        model = model.module
    return model


def save_checkpoint(file, model: nn.Module, optimizer: Union[None, Optimizer] = None, **kwargs):
    checkpoint = {MODEL: get_unwrapped_model(model).state_dict()}
    if optimizer:
        checkpoint[OPTIMIZER] = optimizer.state_dict()
    for k, v in kwargs.items():
        if k in checkpoint:
            raise KeyError(f'Get duplicated key {k}.')
        checkpoint[k] = v
    file_path = os.path.split(file)[0]
    if not os.path.exists(file_path):
        os.makedirs(file_path, exist_ok=True)
    torch.save(checkpoint, file)


def load_checkpoint(file, model: Union[nn.Module, None] = None, optimizer: Union[Optimizer, None] = None):
    logger.info(f'Loading checkpoint {file}.')
    checkpoint = torch.load(file)
    if model:
        if MODEL not in checkpoint:
            logger.warning(f'Checkpoint {file} does not include model state dict.')
        else:
            model = get_unwrapped_model(model)
            model.load_state_dict(checkpoint[MODEL])
            logger.info('Model state loaded.')

    if optimizer:
        if OPTIMIZER not in checkpoint:
            logger.warning(f'Checkpoint {file} does not include optimizer state dict.')
        else:
            optimizer.load_state_dict(checkpoint[OPTIMIZER])
            logger.info(f'Optimizer state loaded.')
    return checkpoint



from typing import Union

from torch.nn.functional import interpolate

from jbag.transforms import Transform


class DownsampleTransform(Transform):
    def __init__(self, keys, scales: Union[list, tuple]):
        """
        Down sample transform.
        Args:
            keys (str or sequence):
            scales (list, tuple): The element type should be float (used for all spatial dimensions) or list or tuple (each spatial dimension has a specific scale)
        """
        super().__init__(keys)
        self.scales = scales

    def _call_fun(self, data):
        for key in self.keys:
            value = data[key]
            results = []
            for scale in self.scales:
                if not isinstance(scale, (list, tuple)):
                    scale = [scale] * (value.ndim - 1)
                assert len(scale) == value.ndim - 1

                if all([i == 1 for i in scale]):
                    results.append(value)
                else:
                    new_shape = [round(i * j) for i, j in zip(value.shape[1:], scale)]
                    results.append(interpolate(value[None].float(), new_shape, mode='nearest-exact')[0].to(value.dtype))
            data[key] = results
        return data


if __name__ == '__main__':
    from cavass.ops import read_cavass_file
    import torch

    image = read_cavass_file('/data1/dj/data/bca/cavass_data/OAM/N007PETCT.BIM')
    image = image[None]
    image = torch.from_numpy(image)
    data = {'image': image}

    t = DownsampleTransform(keys=['image'], scales=[0.5, 0.2])

    data = t(data)

    image = data['image_downsampled']
    image1 = image[0][0].numpy()
    image2 = image[1][0].numpy()



import numpy as np
import torch
from einops import rearrange, repeat

from jbag.transforms.transforms import Transform


class ToType(Transform):
    def __init__(self, keys, dtype):
        super().__init__(keys)
        self.dtype = dtype

    def _call_fun(self, data):
        for key in self.keys:
            value = data[key].astype(self.dtype)
            data[key] = value
        return data


class ToTensor(Transform):
    def __init__(self, keys):
        super().__init__(keys)

    def _call_fun(self, data):
        for key in self.keys:
            data[key] = torch.from_numpy(data[key])
        return data


class Rearrange(Transform):
    def __init__(self, keys, pattern):
        """
        Change the arrangement of given elements.

        Args:
            keys (str or sequence):
            pattern (str): Arranging pattern. For example "i j k -> j k i".
        """
        super().__init__(keys)
        self.pattern = pattern

    def _call_fun(self, data):
        for key in self.keys:
            value = data[key]
            value = rearrange(value, self.pattern)
            data[key] = value
        return data


class Repeat(Transform):
    def __init__(self, keys, pattern, **kwargs):
        super().__init__(keys)
        self.pattern = pattern
        self.kwargs = kwargs

    def _call_fun(self, data):
        for key in self.keys:
            value = data[key]
            value = repeat(value, self.pattern, **self.kwargs)
            data[key] = value
        return data


class AddChannel(Transform):
    def __init__(self, keys, dim):
        """
        Add additional dimension in specific position.

        Args:
            keys (str or sequence):
            dim (int):
        """
        super().__init__(keys)
        self.dim = dim

    def _call_fun(self, data):
        for key in self.keys:
            value = data[key]
            value = np.expand_dims(value, axis=self.dim)
            data[key] = value
        return data
